import 'dart:async';
import 'dart:convert';

import 'package:uuid/uuid.dart';

import '../chatroom_manager.dart';
import 'net_service.dart';

enum AsyncStatus { none, invoking, success, error }

enum MessageType { none, normal, memberLeave }

class Member {
  // consts fields
  static const MEMBERKEYKEY = "memberkey";
  static const NICKNAMEKEY = "nickname";
  static const AVATARKEY = "avatar";

  late String memberKey;
  late String nickName;
  late String avatar;

  // constructors
  Member()
      : memberKey = '',
        nickName = '',
        avatar = '';

  Member.make(
      {required this.memberKey, required this.nickName, required this.avatar});

  Member.fromMap(Map map) {
    memberKey = map[MEMBERKEYKEY];
    nickName = map[NICKNAMEKEY];
    avatar = map[AVATARKEY];
  }

  Map asMap() {
    return {
      MEMBERKEYKEY: memberKey,
      NICKNAMEKEY: nickName,
      AVATARKEY: avatar,
    };
  }
}

class Message {
  // consts fields
  static const MESSAGEIDKEY = "messageid";
  static const SENDERKEY = "sender";
  static const PAYLOADKEY = "payload";

  late String messageId;
  late Member sender;
  late String payload;

  Message.make(
      {required this.messageId, required this.sender, required this.payload});

  Message.fromMap(Map map) {
    messageId = map[MESSAGEIDKEY];
    sender = Member.fromMap(map[SENDERKEY]);
    payload = map[PAYLOADKEY];
  }

  Map asMap() {
    return {
      MESSAGEIDKEY: messageId,
      SENDERKEY: sender.asMap(),
      PAYLOADKEY: payload,
    };
  }
}

class Room {
  // consts fields
  static const ROOMIDKEY = "roomid";
  static const NAMEKEY = "name";
  static const CREATETSKEY = "createts";
  static const MAXMEMBERSKEY = "maxmembers";
  static const MAXMESSAGESKEY = "maxmessages";
  static const PASSWORDKEY = "password";
  static const MEMBERSKEY = "members";
  static const MESSAGESKEY = "messages";

  late String roomId;
  late String name;
  late int createTs;
  late int maxMembers;
  late int maxMessages;
  late String password;
  late List<Member> members;
  late List<Message> messages;
  late ChannelService socks;
  bool connected = false;
  StreamController<int> messageStreamController = StreamController<int>();

  Room.make(
      {required this.roomId,
      required this.name,
      required this.createTs,
      required this.maxMembers,
      required this.maxMessages,
      required this.password,
      required this.members,
      required this.messages});

  Room.fromMap(Map map) {
    roomId = map[ROOMIDKEY];
    name = map[NAMEKEY];
    createTs = map[CREATETSKEY];
    maxMembers = map[MAXMEMBERSKEY];
    maxMessages = map[MAXMESSAGESKEY];
    password = map[PASSWORDKEY];
    members = <Member>[];
    messages = <Message>[];
    for (final map in map[MEMBERSKEY]) {
      members.add(Member.fromMap(map));
    }
  }

  Map asMap() {
    final map = {
      ROOMIDKEY: roomId,
      NAMEKEY: name,
      CREATETSKEY: createTs,
      MAXMEMBERSKEY: maxMembers,
      MAXMESSAGESKEY: maxMessages,
      PASSWORDKEY: password,
    };
    final membersList = <Map>[];
    for (final item in members) {
      membersList.add(item.asMap());
    }
    map.addEntries([MapEntry(MEMBERSKEY, membersList)]);
    return map;
  }

  // TODO: --------------------- 模型逻辑 ------------------------
  ///
  /// 发一个聊天消息
  ///
  Stream<AsyncStatus> sendMessage(Message message) async* {
    socks.sendMessage(jsonEncode(message.asMap()));
    yield AsyncStatus.success;
  }

  ///
  /// 接收到一个聊天逻辑
  ///
  Future revieveAMessage(Message message) async {}

  ///
  /// 获取成员列表
  ///
  Future<List<Member>> memberList() async {
    if (members.isEmpty) {
      // TODO: 读取数据库加载
    }
    return members;
  }

  ///
  /// 获取消息列表
  ///
  Future<List<Message>> messageList() async {
    return messages;
  }

  ///
  /// 连接聊天服务器
  ///
  void joinThisRoom(String me) {
    socks = ChannelService(roomId, me);
    connected = true;
    socks.lookingForMessage((String msgStr) {
      // TODO: 获取到了新的消息
      final map = jsonDecode(msgStr);
      if (map.containsKey('error')) {
        messageStreamController.sink.add(0);
        return leaveThisRoom();
      }
      final ev = MessageEvent.fromMap(map);
      messages.add(ev.message);
      messageStreamController.sink.add(1);
    });
  }

  void leaveThisRoom() {
    socks.channel.sink.close();
    connected = false;
  }
}

class Chat {
  // consts fields
  static const ROMMSKEY = "rooms";

  late List<Room> rooms;
  late StorageBase storage;
  late MessageBase channel;

  Chat.make(
      {required this.rooms, required this.storage, required this.channel});

  Chat.fromMap(Map map) {
    rooms = <Room>[];
    for (final map in map[ROMMSKEY]) {
      rooms.add(Room.fromMap(map));
    }
  }

  Map asMap() {
    final rommsList = <Map>[];
    for (final item in rooms) {
      rommsList.add(item.asMap());
    }
    return {'rooms': rommsList};
  }

  // TODO: --------------------- 模型逻辑 ------------------------
  ///
  /// 本地创建一个聊天室
  ///
  Stream<AsyncStatus> createRoom(
      NewRoomEvent event, List<Member> members) async* {
    rooms.add(event.room);
    event.room.members.add(event.member);
    members.forEach((element) {
      event.room.members.add(element);
    });
    yield AsyncStatus.invoking;
    // TODO: 数据库持久化的逻辑以及网络请求发消息的请求
    final r = await HttpEndpoint('https://api.codiario.com/room/create').post(
        jsonEncode({
      'roomId': event.room.roomId,
      'max': event.room.maxMembers,
      'expire': 180
    }));
    if (r) {
      await _saveRooms();
      await channel.sendMessage(event);
      yield AsyncStatus.success;
    } else {
      rooms = rooms.where((room) => room.roomId != event.room.roomId).toList();
      yield AsyncStatus.error;
    }
  }

  ///
  ///  被邀请进入一个聊天室
  ///
  Future BeInvitedToARoom(NewRoomEvent event, Member me) async {
    rooms.add(event.room);
    event.room.members.add(event.member);
    event.room.members.add(me);
    // TODO: 执行数据库持久化
    await _saveRooms();
  }

  Future _saveRooms() async {
    final savedStr = jsonEncode(rooms.map((e) => e.asMap()).toList());
    await storage.saveRooms(savedStr);
  }

  void reloadRooms() {
    final dataStr = storage.reloadRooms();
    final list = jsonDecode(dataStr);
    for (var map in list) {
      rooms.add(Room.fromMap(map));
    }
  }
}

class NewRoomEvent {
  // consts fields
  static const ROOMKEY = "room";
  static const MEMBERKEY = "member";

  late Room room;
  late Member member;

  NewRoomEvent.make({required this.room, required this.member});
  NewRoomEvent.fromForm(String name, int maxMembers, int maxMessages,
      String password, Member creator) {
    room = Room.make(
        roomId: Uuid().v4(),
        name: name,
        createTs: DateTime.now().millisecondsSinceEpoch,
        maxMembers: maxMembers,
        maxMessages: maxMessages,
        password: password,
        members: [creator],
        messages: []);
    member = creator;
  }

  NewRoomEvent.fromMap(Map map) {
    room = Room.fromMap(map[ROOMKEY]);
    member = Member.fromMap(map[MEMBERKEY]);
  }

  Map asMap() {
    return {ROOMKEY: room.asMap(), MEMBERKEY: member.asMap()};
  }
}

class LeaveEvent {
  // consts fields
  static const ROOMIDKEY = "roomid";
  static const MEMBERPUBKEYKEY = "memberpubkey";

  late String roomId;
  late String memberPubkey;

  LeaveEvent.make({required this.roomId, required this.memberPubkey});

  LeaveEvent.fromMap(Map map) {
    roomId = map[ROOMIDKEY];
    memberPubkey = map[MEMBERPUBKEYKEY];
  }

  Map asMap() {
    return {
      ROOMIDKEY: roomId,
      MEMBERPUBKEYKEY: memberPubkey,
    };
  }
}

class MessageEvent {
  // consts fields
  static const MESSAGEKEY = "message";
  static const LEAVEKEY = "leave";
  static const TYPEKEY = "type";

  late Message message;
  late LeaveEvent leave;
  late int messageType;

  MessageEvent.fromMap(Map map) {
    if (map.containsKey(MESSAGEKEY)) {
      message = Message.fromMap(map[MESSAGEKEY]);
    }
    if (map.containsKey(LEAVEKEY)) {
      leave = LeaveEvent.fromMap(map[LEAVEKEY]);
    }
    messageType = map[TYPEKEY];
  }

  Map asMap() {
    return {
      MESSAGEKEY: message.asMap(),
    };
  }
}
